-----------------------------------------------------------------------------------------------------------------------
-- MIT License.  See the bottom of this file for details.
-----------------------------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------
-- Cleanup old database
----------------------------------------------------------------------------------------------------
use master;
go
if exists (select * from sys.databases where name = 'AutoTracker')
begin
	alter database AutoTracker set offline with rollback immediate;
	alter database AutoTracker set online with rollback immediate;
	drop database AutoTracker;
end
go

----------------------------------------------------------------------------------------------------
-- Create database.  Adjust paths as needed.
----------------------------------------------------------------------------------------------------
create database AutoTracker
on (name = N'AutoTracker', filename = N'd:\data\data2016\data\AutoTracker.mdf' , size = 10240kb , filegrowth = 10240kb )
log on (name = N'AutoTracker_log', filename = N'd:\data\data2016\log\AutoTracker_log.ldf' , size = 10240kb , filegrowth = 10240kb )
go

use AutoTracker;
go

----------------------------------------------------------------------------------------------------
-- Drop existing tables
----------------------------------------------------------------------------------------------------
if exists
(
	select t.name
	from sys.tables t
	inner join sys.schemas s on s.schema_id = t.schema_id
	where s.name = 'dbo'
	and t.name = 'Customer'
	and t.temporal_type_desc = 'SYSTEM_VERSIONED_TEMPORAL_TABLE'
)
begin
	alter table dbo.Customer set (system_versioning = off);
end

drop table if exists history.CustomerHistory;
drop table if exists dbo.Customer;
drop table if exists history.CustomerArchive;
go

----------------------------------------------------------------------------------------------------
-- Create history schema
----------------------------------------------------------------------------------------------------
drop schema if exists history;
go
create schema history authorization dbo;
go

----------------------------------------------------------------------------------------------------
-- Create tables.
----------------------------------------------------------------------------------------------------
create table history.CustomerHistory
(
	CustomerId int not null,
	FirstName varchar(20) not null,
	LastName varchar(20) not null,
	Address varchar(80) not null,
	City varchar(50) not null,
	State char(2) not null,
	ValidFrom datetime2 not null,
	ValidTo datetime2 not null,
	--WhoChanged sysname not null default (suser_sname())
) with (data_compression = page);
go
create clustered index ix1_CustomerHistory on history.CustomerHistory (ValidTo, ValidFrom);
go
create table dbo.Customer
(
	CustomerId int not null identity(1,1),
	FirstName varchar(20) not null,
	LastName varchar(20) not null,
	Address varchar(80) not null,
	City varchar(50) not null,
	State char(2) not null,
	ValidFrom datetime2 generated always as row start not null,
	ValidTo datetime2 generated always as row end not null,
	period for system_time (ValidFrom, ValidTo),
	constraint pk_CustomerId primary key clustered (CustomerId)
)
with (system_versioning = on (history_table = history.CustomerHistory));
go
create table history.CustomerArchive
(
	CustomerId int not null,
	FirstName varchar(20) not null,
	LastName varchar(20) not null,
	Address varchar(80) not null,
	City varchar(50) not null,
	State char(2) not null,
	ValidFrom datetime2 not null,
	ValidTo datetime2 not null,
);
go

----------------------------------------------------------------------------------------------------
-- Add a few records to main table.
----------------------------------------------------------------------------------------------------
insert dbo.Customer (FirstName, LastName, Address, City, State)
values
('Anna', 'Acuff', '1190 E Baxters Ln', 'Gilroy', 'CA'),
('Sharla', 'Scruggs', '11058 W Fennes Rd', 'Buffalo', 'MN'),
('Marcia', 'Blint', '6817 SW Old Orchard Ln', 'Antioch', 'CA'),
('Donna', 'Bishop', '520 W Stoneham Ln', 'North Las Vegas', 'NV'),
('Robert', 'Bove', '5339 N Roberton St', 'Buckeye', 'AZ');

----------------------------------------------------------------------------------------------------
-- Do a bunch of updates to these records. (Bump the street number each time.)
----------------------------------------------------------------------------------------------------
declare @ctr int = 0;
while @ctr < 5000
begin
	with Cust as
	(
		select top (1) Address,
			cast(cast(left(Address, charindex(' ', Address) - 1) as int) + 1 as varchar(20)) + ' ' +
				substring(Address, charindex(' ', Address) + 1, len(Address) - charindex(' ', Address)) NewAddress
		from dbo.Customer
		order by newid()
	)
	update Cust
	set Address = NewAddress;

	select @ctr += 1;
end
go

-- select * from history.CustomerHistory;

----------------------------------------------------------------------------------------------------
-- Archive the data 1000 rows at a time.
-- The following block may be executed a number of times if desired.
-- The idea here is to
--   (1) start a transaction
--   (2) disabled system versioning (since this is happening inside a transaction, this prevents
--       other threads from modifying the temporal table until the transaction is complete)
--   (3) copy the records to the archive table
--   (4) delete the records from the history table
--   (5) re-enabled system versioning
--   (6) commit the transaction
-- We have to use dynamic SQL (or similar mechanisms) to cause these steps to happen in
-- separate batches, otherwise SQL will generate an error related to not being allowed to
-- delete from a history table.
----------------------------------------------------------------------------------------------------

select 'Before running archive operation' Status,
	(select count(*) NbrRows from history.CustomerHistory) HistoryCount,
	(select count(*) NbrRows from history.CustomerArchive) ArchiveCount;

declare @cutoff datetime2;

select @cutoff = ValidTo
from history.CustomerHistory
-- we would probably want to limit the archival process to some threshold date
-- where ValidTo < 'some cutoff date'
order by ValidTo
offset 999 rows
fetch next 1 row only;

declare @offSql nvarchar(max) = 'alter table dbo.Customer set (system_versioning = off);';

declare @insertSql nvarchar(max) = 'insert history.CustomerArchive (CustomerId, FirstName, LastName, Address, City, State, ValidFrom, ValidTo)
	select CustomerId, FirstName, LastName, Address, City, State, ValidFrom, ValidTo
	from history.CustomerHistory
	where ValidTo <= ''' + cast(@cutoff as nvarchar(30)) + ''';';

declare @deleteSql nvarchar(max) = 'delete history.CustomerHistory where ValidTo <= ''' + cast(@cutoff as nvarchar(30)) + ''';';

declare @onSql nvarchar(max) = 'alter table dbo.Customer set (system_versioning = on (history_table = history.CustomerHistory, data_consistency_check = off));';

begin transaction;
	exec (@offSql);
	exec (@insertSql);
	exec (@deleteSql);
	exec (@onSql);
commit transaction;
go

-- Check if it worked
select 'After running archive operation' Status,
	(select count(*) NbrRows from history.CustomerHistory) HistoryCount,
	(select count(*) NbrRows from history.CustomerArchive) ArchiveCount;

-----------------------------------------------------------------------------------------------------------------------
-- Copyright 2017-2019, Allison Benneth (allison at sqltran.org).

-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
-- 
-- The above copyright notice and this permission notice shall be included in all
-- copies or substantial portions of the Software.
-- 
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
-----------------------------------------------------------------------------------------------------------------------
